import typing as t
from abc import ABCMeta, abstractmethod
from .types import SegmentsDict, JSONSignatureDict
from ..errors import InvalidKeyTypeError
from ..registry import Header


class HeaderMember:
    """A header member of the JSON signature. It is combined with protected header,
    and unprotected header.
    """
    def __init__(self, protected: t.Optional[Header] = None, header: t.Optional[Header] = None):
        #: protected header
        self.protected = protected
        #: unprotected header
        self.header = header

    def headers(self) -> Header:
        rv: Header = {}
        if self.protected:
            rv.update(self.protected)
        if self.header:
            rv.update(self.header)
        return rv

    def set_kid(self, kid: str) -> None:
        if self.header is None:
            self.header = {}
        self.header["kid"] = kid


class CompactSignature:
    """JSON Web Signature object for compact mode. This object is used to
    represent the JWS instance.
    """
    def __init__(self, protected: Header, payload: bytes):
        self.protected = protected
        self.payload = payload
        self.segments: SegmentsDict = {}

    def headers(self) -> Header:
        return self.protected

    def set_kid(self, kid: str) -> None:
        self.protected["kid"] = kid


class FlattenedJSONSignature:
    """"JSON Signature object that represents a flattened JSON serialization."""

    #: mark it as flattened
    flattened: t.ClassVar[bool] = True

    def __init__(self, member: HeaderMember, payload: bytes):
        #: the only header member
        self.member = member
        #: payload content
        self.payload = payload
        self.signature: t.Optional[JSONSignatureDict] = None
        self.segments: SegmentsDict = {}

    @property
    def members(self) -> t.List[HeaderMember]:
        return [self.member]

    def headers(self) -> Header:
        return self.member.headers()


class GeneralJSONSignature:
    """"JSON Signature object that represents a general JSON serialization."""

    #: mark it as not flattened (general)
    flattened: t.ClassVar[bool] = False

    def __init__(self, members: t.List[HeaderMember], payload: bytes):
        #: a list of header members
        self.members = members
        #: payload content
        self.payload = payload
        self.signatures: t.List[JSONSignatureDict] = []
        self.segments: SegmentsDict = {}


class JWSAlgModel(object, metaclass=ABCMeta):
    """Interface for JWS algorithm. JWA specification (RFC7518) SHOULD
    implement the algorithms for JWS with this base implementation.
    """

    name: str
    description: str
    recommended: bool = False
    key_type = "oct"
    algorithm_type = "JWS"
    algorithm_location = "sig"

    def check_key_type(self, key: t.Any) -> None:
        if key.key_type != self.key_type:
            raise InvalidKeyTypeError(f'Algorithm "{self.name}" requires "{self.key_type}" key')

    @abstractmethod
    def sign(self, msg: bytes, key: t.Any) -> bytes:
        """Sign the text msg with a private/sign key.

        :param msg: message bytes to be signed
        :param key: private key to sign the message
        :return: bytes
        """

    @abstractmethod
    def verify(self, msg: bytes, sig: bytes, key: t.Any) -> bool:
        """Verify the signature of text msg with a public/verify key.

        :param msg: message bytes to be signed
        :param sig: result signature to be compared
        :param key: public key to verify the signature
        :return: boolean
        """



from joserfc import jwt
from joserfc.jwk import OctKey
from tests.base import TestFixture, load_key


class TestJWTFixtures(TestFixture):
    def run_test(self, data):
        header = data["header"]
        algorithms = [header["alg"]]
        if "secret" in data:
            key = OctKey.import_key(data["secret"])
            private_key = key
            public_key = key
        else:
            private_key = load_key(data["private_key"])
            public_key = load_key(data["public_key"])

        claims = data["payload"]
        value = jwt.encode(header, claims, private_key, algorithms=algorithms)
        if "token" in data:
            self.assertEqual(value, data["token"])

        obj = jwt.decode(value, public_key, algorithms=algorithms)
        self.assertEqual(obj.header["typ"], "JWT")
        self.assertEqual(obj.claims, claims)


TestJWTFixtures.load_fixture('jwt_use_jws.json')



import typing as t
from ..rfc7515.json import verify_signature
from ..rfc7515.types import JSONSignatureDict
from ..jws import (
    HeaderDict,
    HeaderMember,
    FlattenedJSONSignature,
    FlattenedJSONSerialization,
    JWSRegistry as _JWSRegistry,
    serialize_json as _serialize_json,
    deserialize_json as _deserialize_json,
)
from ..jwk import KeyFlexible, guess_key
from ..util import (
    to_bytes,
    to_str,
    json_b64encode,
    json_b64decode,
    urlsafe_b64encode,
)
from ..errors import BadSignatureError
from .registry import JWSRegistry


def serialize_json(
        member: HeaderDict,
        payload: t.Union[str, bytes],
        private_key: KeyFlexible,
        algorithms: t.Optional[t.List[str]] = None,
        registry: t.Optional[_JWSRegistry] = None) -> FlattenedJSONSerialization:

    _member = HeaderMember(**member)
    headers = _member.headers()
    if "b64" not in headers:
        return _serialize_json(member, payload, private_key, algorithms, registry)

    if registry is None:
        registry = JWSRegistry(algorithms=algorithms)

    if headers["b64"] is True:
        return _serialize_json(member, payload, private_key, registry=registry)

    registry.check_header(headers)

    key = guess_key(private_key, _member, True)
    key.check_use("sig")
    alg = registry.get_alg(headers["alg"])

    if _member.protected:
        protected_segment = json_b64encode(_member.protected)
    else:
        protected_segment = b""

    signing_input = b".".join([protected_segment, to_bytes(payload)])
    signature = urlsafe_b64encode(alg.sign(signing_input, key))

    rv: FlattenedJSONSerialization = {
        "payload": to_str(payload),
        "signature": to_str(signature),
    }
    if protected_segment:
        rv["protected"] = to_str(protected_segment)
    if _member.header:
        rv["header"] = _member.header
    return rv


def deserialize_json(
        value: FlattenedJSONSerialization,
        public_key: KeyFlexible,
        algorithms: t.Optional[t.List[str]] = None,
        registry: t.Optional[_JWSRegistry] = None) -> FlattenedJSONSignature:
    obj = _extract_json(value)
    if obj is None:
        return _deserialize_json(value, public_key, algorithms, registry)

    if registry is None:
        registry = JWSRegistry(algorithms=algorithms)

    headers = obj.headers()
    if headers["b64"] is True:
        return _deserialize_json(value, public_key, registry=registry)

    payload_segment = obj.segments["payload"]
    find_key = lambda d: guess_key(public_key, d)
    assert obj.signature is not None
    if not verify_signature(obj.member, obj.signature, payload_segment, registry, find_key):
        raise BadSignatureError()
    return obj


def _extract_json(value: FlattenedJSONSerialization) -> t.Optional[FlattenedJSONSignature]:
    if "signatures" in value:
        return None

    if "protected" in value:
        protected_segment = to_bytes(value["protected"])
        protected = json_b64decode(protected_segment)
    else:
        protected = None

    header = value.get("header")
    member = HeaderMember(protected, header)
    headers = member.headers()
    if "b64" not in headers:
        return None

    payload = to_bytes(value["payload"])
    obj = FlattenedJSONSignature(member, payload)
    _sig: JSONSignatureDict = {"signature": value["signature"]}
    if "protected" in value:
        _sig["protected"] = value["protected"]
    if "header" in value:
        _sig["header"] = value["header"]
    obj.signature = _sig
    obj.segments = {"payload": payload}
    return obj



from cryptography.hazmat.primitives.asymmetric.ec import SECP256K1
from ..rfc7518.ec_key import DSS_CURVES, CURVES_DSS
from ..rfc7518.jws_algs import ECAlgModel

ES256K = ECAlgModel("ES256K", "secp256k1", 256)


def register_secp256k1() -> None:
    # https://tools.ietf.org/html/rfc8812#section-3.1
    DSS_CURVES["secp256k1"] = SECP256K1
    CURVES_DSS[SECP256K1.name] = "secp256k1"


__all__ = ["ES256K", "register_secp256k1"]



import typing as t
from functools import cached_property
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey, Ed25519PrivateKey
from cryptography.hazmat.primitives.asymmetric.ed448 import Ed448PublicKey, Ed448PrivateKey
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PublicKey, X25519PrivateKey
from cryptography.hazmat.primitives.asymmetric.x448 import X448PublicKey, X448PrivateKey
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    PublicFormat,
    PrivateFormat,
    NoEncryption,
)
from ..rfc7517.models import CurveKey
from ..rfc7517.types import KeyParameters
from ..rfc7517.pem import CryptographyBinding
from ..errors import InvalidExchangeKeyError
from ..util import to_bytes, urlsafe_b64decode, urlsafe_b64encode
from ..registry import KeyParameter

PublicOKPKey = t.Union[Ed25519PublicKey, Ed448PublicKey, X25519PublicKey, X448PublicKey]
PrivateOKPKey = t.Union[Ed25519PrivateKey, Ed448PrivateKey, X25519PrivateKey, X448PrivateKey]
OKPDictKey = t.TypedDict("OKPDictKey", {
    "crv": t.Literal["Ed25519", "Ed448", "X25519", "X448"],
    "x": str,
    "d": str,
}, total=False)
PUBLIC_KEYS_MAP: t.Dict[str, t.Type[PublicOKPKey]] = {
    "Ed25519": Ed25519PublicKey,
    "Ed448": Ed448PublicKey,
    "X25519": X25519PublicKey,
    "X448": X448PublicKey,
}
PRIVATE_KEYS_MAP: t.Dict[str, t.Type[PrivateOKPKey]] = {
    "Ed25519": Ed25519PrivateKey,
    "Ed448": Ed448PrivateKey,
    "X25519": X25519PrivateKey,
    "X448": X448PrivateKey,
}
PrivateKeyTypes = (Ed25519PrivateKey, Ed448PrivateKey, X25519PrivateKey, X448PrivateKey)


class OKPBinding(CryptographyBinding):
    ssh_type = b"ssh-ed25519"

    @staticmethod
    def import_private_key(obj: OKPDictKey) -> PrivateOKPKey:
        crv_key: t.Type[PrivateOKPKey] = PRIVATE_KEYS_MAP[obj["crv"]]
        d = urlsafe_b64decode(to_bytes(obj["d"]))
        return crv_key.from_private_bytes(d)

    @staticmethod
    def import_public_key(obj: OKPDictKey) -> PublicOKPKey:
        crv_key: t.Type[PublicOKPKey] = PUBLIC_KEYS_MAP[obj["crv"]]
        x_bytes = urlsafe_b64decode(to_bytes(obj["x"]))
        return crv_key.from_public_bytes(x_bytes)

    @staticmethod
    def export_private_key(key: PrivateOKPKey) -> t.Dict[str, str]:
        obj = OKPBinding.export_public_key(key.public_key())
        d_bytes = key.private_bytes(Encoding.Raw, PrivateFormat.Raw, NoEncryption())
        obj["d"] = urlsafe_b64encode(d_bytes).decode("utf-8")
        return obj

    @staticmethod
    def export_public_key(key: PublicOKPKey) -> t.Dict[str, str]:
        x_bytes = key.public_bytes(Encoding.Raw, PublicFormat.Raw)
        return {
            "crv": get_key_curve(key),
            "x": urlsafe_b64encode(x_bytes).decode("utf-8"),
        }


class OKPKey(CurveKey[PrivateOKPKey, PublicOKPKey]):
    """Key class of the ``OKP`` key type."""

    key_type = "OKP"
    #: Registry definition for OKP Key
    #: https://www.rfc-editor.org/rfc/rfc8037#section-2
    value_registry = {
        "crv": KeyParameter("Curve", "str", private=False, required=True),
        "x": KeyParameter("X Coordinate", "str", private=False, required=True),
        "d": KeyParameter("OKP Private Key", "str", private=True, required=False),
    }
    binding = OKPBinding
    required_fields = frozenset(["crv", "x"])
    private_only_fields = frozenset(["d"])

    def exchange_derive_key(self, key: "OKPKey") -> bytes:
        # used in ECDH-ES Algorithms
        pubkey: t.Union[X25519PublicKey, X448PublicKey] = key.get_op_key("deriveKey")  # type: ignore[assignment]
        if isinstance(self.private_key, X25519PrivateKey) and isinstance(pubkey, X25519PublicKey):
            return self.private_key.exchange(pubkey)
        elif isinstance(self.private_key, X448PrivateKey) and isinstance(pubkey, X448PublicKey):
            return self.private_key.exchange(pubkey)
        raise InvalidExchangeKeyError()

    @property
    def is_private(self) -> bool:
        return isinstance(self.raw_value, PrivateKeyTypes)

    @cached_property
    def public_key(self) -> PublicOKPKey:
        if isinstance(self.raw_value, PrivateKeyTypes):
            return self.raw_value.public_key()
        return self.raw_value

    @property
    def private_key(self) -> t.Optional[PrivateOKPKey]:
        if isinstance(self.raw_value, PrivateKeyTypes):
            return self.raw_value
        return None

    @property
    def curve_name(self) -> str:
        return get_key_curve(self.raw_value)

    @classmethod
    def generate_key(
            cls,
            crv: str = "Ed25519",
            parameters: t.Optional[KeyParameters] = None,
            private: bool = True,
            auto_kid: bool = False) -> "OKPKey":
        """Generate a ``OKPKey`` with the given "crv" value.

        :param crv: OKPKey curve name
        :param parameters: extra parameter in JWK
        :param private: generate a private key or public key
        :param auto_kid: add ``kid`` automatically
        """
        if crv not in PRIVATE_KEYS_MAP:
            raise ValueError('Invalid crv value: "{}"'.format(crv))

        private_key_cls: t.Type[PrivateOKPKey] = PRIVATE_KEYS_MAP[crv]
        raw_key = private_key_cls.generate()
        if private:
            key = cls(raw_key, raw_key, parameters)
        else:
            pub_key = raw_key.public_key()
            key = cls(pub_key, pub_key, parameters)
        if auto_kid:
            key.ensure_kid()
        return key


def get_key_curve(key: t.Union[PublicOKPKey, PrivateOKPKey]) -> str:
    if isinstance(key, (Ed25519PublicKey, Ed25519PrivateKey)):
        return "Ed25519"
    elif isinstance(key, (Ed448PublicKey, Ed448PrivateKey)):
        return "Ed448"
    elif isinstance(key, (X25519PublicKey, X25519PrivateKey)):
        return "X25519"
    elif isinstance(key, (X448PublicKey, X448PrivateKey)):
        return "X448"
    raise ValueError("Invalid key")  # pragma: no cover



import zlib
from typing import List
from ..rfc7516.models import JWEZipModel


class DeflateZipModel(JWEZipModel):
    name = "DEF"
    description = "DEFLATE"

    def compress(self, s: bytes) -> bytes:
        """Compress bytes data with DEFLATE algorithm."""
        data = zlib.compress(s)
        # drop gzip headers and tail
        return data[2:-4]

    def decompress(self, s: bytes) -> bytes:
        """Decompress DEFLATE bytes data."""
        return zlib.decompress(s, -zlib.MAX_WBITS)


JWE_ZIP_MODELS: List[JWEZipModel] = [DeflateZipModel()]



from unittest import TestCase
from joserfc.jwk import OKPKey
from joserfc.errors import InvalidExchangeKeyError
from tests.keys import read_key


class TestOKPKey(TestCase):
    def test_exchange_derive_key(self):
        key1 = OKPKey.generate_key("Ed448")
        key2 = OKPKey.generate_key("Ed448")
        # only X25519 and X448 can exchange
        self.assertRaises(InvalidExchangeKeyError, key1.exchange_derive_key, key2)

        key3 = OKPKey.generate_key("X25519", private=False)
        key4 = OKPKey.generate_key("X25519")
        # key3 must have private key
        self.assertRaises(InvalidExchangeKeyError, key3.exchange_derive_key, key4)

        key5 = OKPKey.generate_key("X448")
        # curve name doesn't match
        self.assertRaises(InvalidExchangeKeyError, key4.exchange_derive_key, key5)

    def test_generate_keys(self):
        curves = ["Ed25519", "Ed448", "X25519", "X448"]
        for crv in curves:
            key = OKPKey.generate_key(crv)
            self.assertTrue(key.is_private)
            self.assertEqual(key.curve_name, crv)

        public_key = OKPKey.generate_key("Ed25519", private=False)
        self.assertFalse(public_key.is_private)
        self.assertIsNone(public_key.kid)
        self.assertRaises(ValueError, OKPKey.generate_key, "invalid")

        key = OKPKey.generate_key(auto_kid=True)
        self.assertIsNotNone(key.kid)

    def test_import_pem_key(self):
        private_pem = read_key("okp-ed448-private.pem")
        public_pem = read_key("okp-ed448-public.pem")
        private_key: OKPKey = OKPKey.import_key(private_pem)
        public_key: OKPKey = OKPKey.import_key(public_pem)

        self.assertEqual(private_key.as_pem(), private_pem)
        self.assertEqual(private_key.as_pem(private=False), public_pem)
        self.assertEqual(public_key.as_pem(), public_pem)

        self.assertIn("d", private_key.as_dict())
        self.assertNotIn("d", public_key.as_dict())

    def test_properties(self):
        private_pem = read_key("okp-ed448-private.pem")
        public_pem = read_key("okp-ed448-public.pem")
        private_key: OKPKey = OKPKey.import_key(private_pem)
        public_key: OKPKey = OKPKey.import_key(public_pem)

        self.assertTrue(private_key.is_private)
        self.assertFalse(public_key.is_private)

        self.assertEqual(private_key.private_key, private_key.raw_value)
        self.assertEqual(public_key.public_key, public_key.raw_value)
        self.assertIsNone(public_key.private_key)

    def test_import_from_json(self):
        private_key = OKPKey.import_key(read_key("okp-ed25519-private.json"))
        public_key = OKPKey.import_key(read_key("okp-ed25519-public.json"))
        self.assertTrue(private_key.is_private)
        self.assertFalse(public_key.is_private)

    def test_all_as_methods(self):
        private_json = read_key("okp-ed25519-private.json")
        public_json = read_key("okp-ed25519-public.json")
        key: OKPKey = OKPKey.import_key(private_json)

        # as_dict
        data = key.as_dict()
        self.assertIn('d', data)
        self.assertEqual(data, private_json)
        data = key.as_dict(private=False)
        self.assertNotIn('d', data)
        self.assertEqual(data, public_json)

        # as_pem
        data = key.as_pem()
        self.assertIn(b"PRIVATE", data)
        data = key.as_pem(private=False)
        self.assertIn(b"PUBLIC", data)

        # as_der
        data = key.as_der()
        self.assertIsInstance(data, bytes)

    def test_output_with_password(self):
        key = OKPKey.import_key(read_key("okp-ed25519-private.json"))
        pem = key.as_pem(password="secret")
        self.assertRaises(
            TypeError,
            OKPKey.import_key,
            pem
        )
        key2 = OKPKey.import_key(pem, password="secret")
        self.assertEqual(key.as_pem(), key2.as_pem())



from .models import CompactEncryption, Recipient
from .._keys import Key
from ..errors import (
    MissingAlgorithmError,
    MissingEncryptionError,
    DecodeError,
)
from ..util import (
    json_b64decode,
    urlsafe_b64encode,
    urlsafe_b64decode,
)


def represent_compact(obj: CompactEncryption) -> bytes:
    assert obj.recipient is not None
    encrypted_key = obj.recipient.encrypted_key
    assert encrypted_key is not None
    return b".".join([
        obj.base64_segments["aad"],
        urlsafe_b64encode(encrypted_key),
        obj.base64_segments["iv"],
        obj.base64_segments["ciphertext"],
        obj.base64_segments["tag"],
    ])


def extract_compact(value: bytes) -> CompactEncryption:
    parts = value.split(b".")
    if len(parts) != 5:
        raise ValueError("Invalid JSON Web Encryption")

    header_segment, ek_segment, iv_segment, ciphertext_segment, tag_segment = parts
    try:
        protected = json_b64decode(header_segment)
        if "alg" not in protected:
            raise MissingAlgorithmError()
        if "enc" not in protected:
            raise MissingEncryptionError()
    except (TypeError, ValueError):
        raise DecodeError("Invalid header")

    obj = CompactEncryption(protected)
    obj.base64_segments.update({
        "aad": header_segment,
        "iv": iv_segment,
        "ciphertext": ciphertext_segment,
        "tag": tag_segment,
    })
    obj.bytes_segments.update({
        "iv": urlsafe_b64decode(iv_segment),
        "ciphertext": urlsafe_b64decode(ciphertext_segment),
        "tag": urlsafe_b64decode(tag_segment),
    })
    recipient: Recipient[Key] = Recipient(obj)
    recipient.encrypted_key = urlsafe_b64decode(ek_segment)
    obj.recipient = recipient
    return obj



import os
import typing as t
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.keywrap import (
    aes_key_wrap,
    aes_key_unwrap,
    InvalidUnwrap,
)
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import GCM
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.exceptions import InvalidTag
from .derive_key import derive_key_for_concat_kdf
from .oct_key import OctKey
from .rsa_key import RSAKey
from .ec_key import ECKey
from ..rfc7516.models import (
    JWEAlgModel,
    JWEDirectEncryption,
    JWEKeyEncryption,
    JWEKeyWrapping,
    JWEKeyAgreement,
    JWEEncModel,
    Recipient,
)
from ..util import to_bytes, urlsafe_b64encode, urlsafe_b64decode
from ..registry import HeaderParameter
from ..errors import (
    InvalidKeyLengthError,
    DecodeError,
)


class DirectAlgModel(JWEDirectEncryption):
    name = "dir"
    description = "Direct use of a shared symmetric key"
    recommended = True

    def compute_cek(self, size: int, recipient: Recipient[OctKey]) -> bytes:
        key = recipient.recipient_key
        assert key is not None
        self.check_key_type(key)
        cek = key.raw_value
        if len(cek) * 8 != size:
            raise InvalidKeyLengthError(f"A key of size {size} bits MUST be used")
        return cek


class RSAAlgModel(JWEKeyEncryption):
    #: A key of size 2048 bits or larger MUST be used with these algorithms
    #: RSA1_5, RSA-OAEP, RSA-OAEP-256
    key_size = 2048
    key_types = ["RSA"]

    def __init__(
            self,
            name: str,
            description: str,
            pad_fn: padding.AsymmetricPadding,
            recommended: bool = False):
        self.name = name
        self.description = description
        self.padding = pad_fn
        self.recommended = recommended

    def encrypt_cek(self, cek: bytes, recipient: Recipient[RSAKey]) -> bytes:
        key = recipient.recipient_key
        assert key is not None
        self.check_key_type(key)
        op_key = key.get_op_key("encrypt")
        if op_key.key_size < self.key_size:
            raise InvalidKeyLengthError(f"A key of size {self.key_size} bits or larger MUST be used")
        return op_key.encrypt(cek, self.padding)

    def decrypt_cek(self, recipient: Recipient[RSAKey]) -> bytes:
        key = recipient.recipient_key
        assert key is not None
        self.check_key_type(key)
        op_key = key.get_op_key("decrypt")
        try:
            assert recipient.encrypted_key is not None
            cek = op_key.decrypt(recipient.encrypted_key, self.padding)
        except ValueError as error:
            raise DecodeError(str(error))
        return cek


class AESAlgModel(JWEKeyWrapping):
    def __init__(self, key_size: int, recommended: bool = False):
        self.name = f"A{key_size}KW"
        self.description = f"AES Key Wrap using {key_size}-bit key"
        self.key_size = key_size
        self.recommended = recommended

    def wrap_cek(self, cek: bytes, key: bytes) -> bytes:
        self.check_op_key(key)
        return aes_key_wrap(key, cek, default_backend())

    def unwrap_cek(self, ek: bytes, key: bytes) -> bytes:
        self.check_op_key(key)
        try:
            cek = aes_key_unwrap(key, ek, default_backend())
        except InvalidUnwrap:
            raise DecodeError("Unwrap AES key failed")
        return cek

    def encrypt_cek(self, cek: bytes, recipient: Recipient[OctKey]) -> bytes:
        key = recipient.recipient_key
        assert key is not None
        self.check_key_type(key)
        op_key = key.get_op_key("wrapKey")
        return self.wrap_cek(cek, op_key)

    def decrypt_cek(self, recipient: Recipient[OctKey]) -> bytes:
        key = recipient.recipient_key
        assert key is not None
        self.check_key_type(key)
        op_key = key.get_op_key("unwrapKey")
        assert recipient.encrypted_key is not None
        return self.unwrap_cek(recipient.encrypted_key, op_key)


class AESGCMAlgModel(JWEKeyWrapping):
    more_header_registry = {
        "iv": HeaderParameter("Initialization vector", "str", True),
        "tag": HeaderParameter("Authentication tag", "str", True),
    }

    def __init__(self, key_size: int):
        self.name = f"A{key_size}GCMKW"
        self.description = f"Key wrapping with AES GCM using {key_size}-bit key"
        self.key_size = key_size

    def wrap_cek(self, cek: bytes, key: bytes) -> bytes:  # pragma: no cover
        raise RuntimeError(f"{self.name} can not be used together with Key Agreement")

    def unwrap_cek(self, ek: bytes, key: bytes) -> bytes:  # pragma: no cover
        raise RuntimeError(f"{self.name} can not be used together with Key Agreement")

    def encrypt_cek(self, cek: bytes, recipient: Recipient[OctKey]) -> bytes:
        key = recipient.recipient_key
        assert key is not None
        self.check_key_type(key)
        op_key = key.get_op_key("wrapKey")
        self.check_op_key(op_key)

        #: https://tools.ietf.org/html/rfc7518#section-4.7.1.1
        #: The "iv" (initialization vector) Header Parameter value is the
        #: base64url-encoded representation of the 96-bit IV value
        iv_size = 96
        iv = os.urandom(iv_size // 8)

        cipher = Cipher(AES(op_key), GCM(iv), backend=default_backend())
        enc = cipher.encryptor()

        encrypted_key = enc.update(cek) + enc.finalize()
        recipient.add_header("iv", urlsafe_b64encode(iv).decode("ascii"))
        recipient.add_header("tag", urlsafe_b64encode(enc.tag).decode("ascii"))
        return encrypted_key

    def decrypt_cek(self, recipient: Recipient[OctKey]) -> bytes:
        key = recipient.recipient_key
        assert key is not None
        self.check_key_type(key)
        op_key = key.get_op_key("unwrapKey")
        self.check_op_key(op_key)

        headers = recipient.headers()
        assert "iv" in headers
        assert "tag" in headers
        iv = urlsafe_b64decode(to_bytes(headers["iv"]))
        tag = urlsafe_b64decode(to_bytes(headers["tag"]))

        cipher = Cipher(AES(op_key), GCM(iv, tag), backend=default_backend())
        d = cipher.decryptor()
        try:
            assert recipient.encrypted_key is not None
            cek = d.update(recipient.encrypted_key) + d.finalize()
        except InvalidTag as error:
            raise DecodeError(str(error))
        return cek


class ECDHESAlgModel(JWEKeyAgreement):
    key_types = ["EC", "OKP"]
    more_header_registry = {
        "epk": HeaderParameter("Ephemeral Public Key", "jwk", True),
        "apu": HeaderParameter("Agreement PartyUInfo", "str"),
        "apv": HeaderParameter("Agreement PartyVInfo", "str"),
    }

    # https://tools.ietf.org/html/rfc7518#section-4.6
    def __init__(self, key_wrapping: t.Optional[JWEKeyWrapping] = None):
        if key_wrapping is None:
            self.name = "ECDH-ES"
            self.description = "ECDH-ES in the Direct Key Agreement mode"
            self.key_size = None
            self.recommended = True
        else:
            self.name = f"ECDH-ES+{key_wrapping.name}"
            self.description = f"ECDH-ES using Concat KDF and CEK wrapped with {key_wrapping.name}"
            self.key_size = key_wrapping.key_size
            self.recommended = key_wrapping.recommended
        self.key_wrapping = key_wrapping

    def encrypt_agreed_upon_key(self, enc: JWEEncModel, recipient: Recipient[ECKey]) -> bytes:
        recipient_key = recipient.recipient_key
        assert recipient_key is not None

        ephemeral_key = recipient.ephemeral_key
        assert ephemeral_key is not None

        shared_key = ephemeral_key.exchange_derive_key(recipient_key)
        headers = recipient.headers()
        return derive_key_for_concat_kdf(shared_key, headers, enc.cek_size, self.key_size)

    def decrypt_agreed_upon_key(self, enc: JWEEncModel, recipient: Recipient[ECKey]) -> bytes:
        headers = recipient.headers()
        assert "epk" in headers

        recipient_key = recipient.recipient_key
        assert recipient_key is not None

        self.check_key_type(recipient_key)
        ephemeral_key = recipient_key.import_key(headers["epk"])
        shared_key = recipient_key.exchange_derive_key(ephemeral_key)
        return derive_key_for_concat_kdf(shared_key, headers, enc.cek_size, self.key_size)


class PBES2HSAlgModel(JWEKeyEncryption):
    # https://www.rfc-editor.org/rfc/rfc7518#section-4.8
    key_size: int
    more_header_registry = {
        "p2s": HeaderParameter("PBES2 Salt Input", "str", True),
        "p2c": HeaderParameter("PBES2 Count", "int", True),
    }
    key_types = ["oct"]

    # A minimum iteration count of 1000 is RECOMMENDED.
    DEFAULT_P2C = 2048

    def __init__(self, hash_size: int, key_wrapping: JWEKeyWrapping):
        self.name = f"PBES2-HS{hash_size}+{key_wrapping.name}"
        self.description = f"PBES2 with HMAC SHA-{hash_size} and {key_wrapping.name} wrapping"
        self.key_size = key_wrapping.key_size
        self.key_wrapping = key_wrapping
        self.hash_alg = getattr(hashes, f"SHA{hash_size}")()

    def compute_derived_key(self, key: bytes, p2s: bytes, p2c: int) -> bytes:
        # The salt value used is (UTF8(Alg) || 0x00 || Salt Input)
        salt = to_bytes(self.name) + b"\x00" + p2s
        kdf = PBKDF2HMAC(
            algorithm=self.hash_alg,
            length=self.key_size // 8,
            salt=salt,
            iterations=p2c,
            backend=default_backend(),
        )
        return kdf.derive(key)

    def encrypt_cek(self, cek: bytes, recipient: Recipient[OctKey]) -> bytes:
        headers = recipient.headers()
        if "p2s" not in headers:
            p2s = os.urandom(16)
            recipient.add_header("p2s", urlsafe_b64encode(p2s).decode("ascii"))
        else:
            p2s = urlsafe_b64decode(to_bytes(headers["p2s"]))

        if "p2c" not in headers:
            # A minimum iteration count of 1000 is RECOMMENDED.
            p2c = self.DEFAULT_P2C
            recipient.add_header("p2c", p2c)
        else:
            p2c = headers["p2c"]

        key = recipient.recipient_key
        assert key is not None
        self.check_key_type(key)
        kek = self.compute_derived_key(key.get_op_key("deriveKey"), p2s, p2c)
        return self.key_wrapping.wrap_cek(cek, kek)

    def decrypt_cek(self, recipient: Recipient[OctKey]) -> bytes:
        headers = recipient.headers()
        assert "p2s" in headers
        assert "p2c" in headers
        p2s = urlsafe_b64decode(to_bytes(headers["p2s"]))
        p2c = headers["p2c"]

        key = recipient.recipient_key
        assert key is not None

        self.check_key_type(key)
        kek = self.compute_derived_key(key.get_op_key("deriveKey"), p2s, p2c)
        assert recipient.encrypted_key is not None
        return self.key_wrapping.unwrap_cek(recipient.encrypted_key, kek)


A128KW = AESAlgModel(128, True)  # A128KW, Recommended
A192KW = AESAlgModel(192)  # A192KW
A256KW = AESAlgModel(256, True)  # A256KW, Recommended


#: https://www.rfc-editor.org/rfc/rfc7518#section-4.1
JWE_ALG_MODELS: t.List[JWEAlgModel] = [
    # Avoid all RSA-PKCS1 v1.5 encryption algorithms ([RFC8017], Section 7.2),
    # preferring RSAES-OAEP ([RFC8017], Section 7.1).
    # https://www.rfc-editor.org/rfc/rfc8725#section-3.2
    RSAAlgModel("RSA1_5", "RSAES-PKCS1-v1_5", padding.PKCS1v15()),
    RSAAlgModel(
        "RSA-OAEP",
        "RSAES OAEP using default parameters",
        padding.OAEP(padding.MGF1(hashes.SHA1()), hashes.SHA1(), None),
        True,
    ),  # Recommended+
    RSAAlgModel(
        "RSA-OAEP-256",
        "RSAES OAEP using SHA-256 and MGF1 with SHA-256",
        padding.OAEP(padding.MGF1(hashes.SHA256()), hashes.SHA256(), None),
    ),
    A128KW,
    A192KW,
    A256KW,
    DirectAlgModel(),  # dir, Recommended
    ECDHESAlgModel(None),  # ECDH-ES, Recommended+
    ECDHESAlgModel(A128KW),  # ECDH-ES+A128KW, Recommended
    ECDHESAlgModel(A192KW),  # ECDH-ES+A192KW
    ECDHESAlgModel(A256KW),  # ECDH-ES+A256KW, Recommended
    AESGCMAlgModel(128),  # A128GCMKW
    AESGCMAlgModel(192),  # A192GCMKW
    AESGCMAlgModel(256),  # A256GCMKW
    PBES2HSAlgModel(256, A128KW),  # PBES2-HS256+A128KW
    PBES2HSAlgModel(384, A192KW),  # PBES2-HS384+A192KW
    PBES2HSAlgModel(512, A256KW),  # PBES2-HS512+A256KW
]



from unittest import TestCase
from joserfc import jwe
from joserfc.jwe import GeneralJSONEncryption
from joserfc.jwk import KeySet, RSAKey, ECKey, OctKey
from joserfc.errors import ConflictAlgorithmError


class TestJWEJSON(TestCase):
    def test_multiple_recipients_with_key(self):
        key1 = RSAKey.generate_key()
        key2 = ECKey.generate_key()
        obj = GeneralJSONEncryption({"enc": "A128CBC-HS256"}, b"i")
        obj.add_recipient({"alg": "RSA-OAEP"}, key1)
        obj.add_recipient({"alg": "ECDH-ES+A128KW"}, key2)
        value = jwe.encrypt_json(obj, None)
        self.assertIn("recipients", value)
        self.assertEqual(len(value["recipients"]), 2)

    def test_multiple_recipients_without_key(self):
        key1 = RSAKey.generate_key(parameters={"kid": "rsa"})
        key2 = ECKey.generate_key(parameters={"kid": "ec"})
        obj = GeneralJSONEncryption({"enc": "A128CBC-HS256"}, b"i")
        obj.add_recipient({"alg": "RSA-OAEP", "kid": "rsa"})
        obj.add_recipient({"alg": "ECDH-ES+A128KW", "kid": "ec"})
        value = jwe.encrypt_json(obj, KeySet([key1, key2]))
        self.assertIn("recipients", value)
        self.assertEqual(len(value["recipients"]), 2)

    def test_multiple_recipients_with_direct_mode(self):
        obj = GeneralJSONEncryption({"enc": "A128CBC-HS256"}, b"i")
        obj.add_recipient({"alg": "dir"}, OctKey.generate_key())
        obj.add_recipient({"alg": "RSA-OAEP"}, RSAKey.generate_key())
        self.assertRaises(
            ConflictAlgorithmError,
            jwe.encrypt_json,
            obj, None,
        )

    def test_with_aad(self):
        obj = GeneralJSONEncryption({"enc": "A128CBC-HS256"}, b"i", aad=b"foo")
        key1 = RSAKey.generate_key()
        obj.add_recipient({"alg": "RSA-OAEP"}, key1)
        value = jwe.encrypt_json(obj, None)
        obj1 = jwe.decrypt_json(value, key1)
        self.assertEqual(obj1.aad, b"foo")



from unittest import TestCase
from joserfc.jwk import RSAKey
from tests.keys import read_key


class TestRSAKey(TestCase):
    def test_import_key_from_dict(self):
        # https://www.rfc-editor.org/rfc/rfc7517#appendix-A.1
        data = {
            "kty": "RSA",
            "n": (
                "0vx7agoebGcQSuuPiLJXZptN9nndrQmbXEps2aiAFbWhM78LhWx4cbbfAAtVT86"
                "zwu1RK7aPFFxuhDR1L6tSoc_BJECPebWKRXjBZCiFV4n3oknjhMstn64tZ_2W-5"
                "JsGY4Hc5n9yBXArwl93lqt7_RN5w6Cf0h4QyQ5v-65YGjQR0_FDW2QvzqY368QQ"
                "MicAtaSqzs8KJZgnYb9c7d0zgdAZHzu6qMQvRL5hajrn1n91CbOpbISD08qNLyr"
                "dkt-bFTWhAI4vMQFh6WeZu0fM4lFd2NcRwr3XPksINHaQ-G_xBniIqbw0Ls1jF4"
                "4-csFCur-kEgU8awapJzKnqDKgw"
            ),
            "e": "AQAB",
            "alg": "RS256",
            "kid": "2011-04-29",
        }
        key: RSAKey = RSAKey.import_key(data)
        self.assertEqual(key.as_dict(), data)
        self.assertFalse(key.is_private)
        self.assertIsNone(key.private_key)

    def test_with_oth(self):
        data = {
            "kty": "RSA",
            "n": (
                "0vx7agoebGcQSuuPiLJXZptN9nndrQmbXEps2aiAFbWhM78LhWx4cbbfAAtVT86"
                "zwu1RK7aPFFxuhDR1L6tSoc_BJECPebWKRXjBZCiFV4n3oknjhMstn64tZ_2W-5"
                "JsGY4Hc5n9yBXArwl93lqt7_RN5w6Cf0h4QyQ5v-65YGjQR0_FDW2QvzqY368QQ"
                "MicAtaSqzs8KJZgnYb9c7d0zgdAZHzu6qMQvRL5hajrn1n91CbOpbISD08qNLyr"
                "dkt-bFTWhAI4vMQFh6WeZu0fM4lFd2NcRwr3XPksINHaQ-G_xBniIqbw0Ls1jF4"
                "4-csFCur-kEgU8awapJzKnqDKgw"
            ),
            "e": "AQAB",
            "oth": "invalid information"
        }
        self.assertRaises(
            ValueError,
            RSAKey.import_key, data
        )

    def test_import_only_from_d(self):
        data = {
            "kty": "RSA",
            "n": (
                "oahUIoWw0K0usKNuOR6H4wkf4oBUXHTxRvgb48E-BVvxkeDNjbC4he8rUWcJoZmd"
                "s2h7M70imEVhRU5djINXtqllXI4DFqcI1DgjT9LewND8MW2Krf3Spsk_ZkoFnila"
                "kGygTwpZ3uesH-PFABNIUYpOiN15dsQRkgr0vEhxN92i2asbOenSZeyaxziK72Uw"
                "xrrKoExv6kc5twXTq4h-QChLOln0_mtUZwfsRaMStPs6mS6XrgxnxbWhojf663tu"
                "EQueGC-FCMfra36C9knDFGzKsNa7LZK2djYgyD3JR_MB_4NUJW_TqOQtwHYbxevo"
                "JArm-L5StowjzGy-_bq6Gw"
            ),
            "e": "AQAB",
            "d": (
                "kLdtIj6GbDks_ApCSTYQtelcNttlKiOyPzMrXHeI-yk1F7-kpDxY4-WY5NWV5Knt"
                "aEeXS1j82E375xxhWMHXyvjYecPT9fpwR_M9gV8n9Hrh2anTpTD93Dt62ypW3yDs"
                "JzBnTnrYu1iwWRgBKrEYY46qAZIrA2xAwnm2X7uGR1hghkqDp0Vqj3kbSCz1XyfC"
                "s6_LehBwtxHIyh8Ripy40p24moOAbgxVw3rxT_vlt3UVe4WO3JkJOzlpUf-KTVI2"
                "Ptgm-dARxTEtE-id-4OJr0h-K-VFs3VSndVTIznSxfyrj8ILL6MG_Uv8YAu7VILS"
                "B3lOW085-4qE3DzgrTjgyQ"
            ),
        }
        key: RSAKey = RSAKey.import_key(data)
        self.assertTrue(key.is_private)
        data["p"] = (
            "9gY2w6I6S6L0juEKsbeDAwpd9WMfgqFoeA9vEyEUuk4kLwBKcoe1x4HG68ik918hdDSE"
            "9vDQSccA3xXHOAFOPJ8R9EeIAbTi1VwBYnbTp87X-xcPWlEPkrdoUKW60tgs1aNd_Nnc"
            "9LEVVPMS390zbFxt8TN_biaBgelNgbC95sM"
        )
        self.assertRaises(ValueError, RSAKey.import_key, data)

    def test_import_key_from_ssh(self):
        ssh_public_pem = read_key("ssh-rsa-public.pem")
        key: RSAKey = RSAKey.import_key(ssh_public_pem)
        self.assertFalse(key.is_private)

        ssh_private_pem = read_key("ssh-rsa-private.pem")
        key: RSAKey = RSAKey.import_key(ssh_private_pem)
        self.assertTrue(key.is_private)

    def test_import_key_from_openssl(self):
        public_pem = read_key("rsa-openssl-public.pem")
        key: RSAKey = RSAKey.import_key(public_pem)
        self.assertFalse(key.is_private)

        private_pem = read_key("rsa-openssl-private.pem")
        key: RSAKey = RSAKey.import_key(private_pem)
        self.assertTrue(key.is_private)

    def test_output_as_methods(self):
        private_pem = read_key("rsa-openssl-private.pem")
        key: RSAKey = RSAKey.import_key(private_pem)

        # as_dict
        data = key.as_dict()
        self.assertIn('d', data)
        data = key.as_dict(private=True)
        self.assertIn('d', data)
        data = key.as_dict(private=False)
        self.assertNotIn('d', data)

        # as_pem
        data = key.as_pem()
        self.assertIn(b"PRIVATE", data)
        data = key.as_pem(private=True)
        self.assertIn(b"PRIVATE", data)
        data = key.as_pem(private=False)
        self.assertIn(b"PUBLIC", data)

        # as_der
        data = key.as_der()
        self.assertIsInstance(data, bytes)

    def test_generate_key(self):
        self.assertRaises(ValueError, RSAKey.generate_key, 8)
        self.assertRaises(ValueError, RSAKey.generate_key, 601)

        key: RSAKey = RSAKey.generate_key(private=False)
        self.assertFalse(key.is_private)
        self.assertIsNone(key.kid)

        key = RSAKey.generate_key(auto_kid=True)
        self.assertIsNotNone(key.kid)

    def test_import_from_der_bytes(self):
        origin_key = RSAKey.generate_key()
        value1 = origin_key.as_der()
        key1 = RSAKey.import_key(value1)
        self.assertEqual(value1, key1.as_der())

    def test_output_with_password(self):
        private_pem = read_key("rsa-openssl-private.pem")
        key: RSAKey = RSAKey.import_key(private_pem)
        pem = key.as_pem(password="secret")
        self.assertRaises(
            TypeError,
            RSAKey.import_key,
            pem
        )
        key2 = RSAKey.import_key(pem, password="secret")
        self.assertEqual(key.as_dict(), key2.as_dict())



from unittest import TestCase
from joserfc import jws, jwe, jwt
from joserfc.jwk import OctKey
from joserfc.errors import (
    InvalidPayloadError,
    InvalidTypeError,
    MissingClaimError,
)


class TestJWT(TestCase):
    def test_invalid_payload(self):
        data = jws.serialize_compact({"alg": "HS256"}, b"hello", "secret")
        self.assertRaises(InvalidPayloadError, jwt.decode, data, "secret")

    def test_invalid_type(self):
        data = jws.serialize_compact({"alg": "HS256", "typ": "JOSE"}, b'{"iss":"a"}', "secret")
        self.assertRaises(InvalidTypeError, jwt.decode, data, "secret")

    def test_claims_registry(self):
        data = jwt.encode({"alg": "HS256"}, {"sub": "a"}, "secret")
        token = jwt.decode(data, "secret")

        claims_registry = jwt.JWTClaimsRegistry(iss={"essential": True})
        self.assertRaises(MissingClaimError, claims_registry.validate, token.claims)

        data = jwt.encode({"alg": "HS256"}, {"iss": "a"}, "secret")
        obj = jwt.decode(data, "secret")
        self.assertEqual(obj.claims["iss"], "a")

    def test_jwe_format(self):
        header = {"alg": "A128KW", "enc": "A128GCM"}
        claims = {"iss": "https://authlib.org"}
        key = OctKey.generate_key(128)
        result = jwt.encode(header, claims, key)
        self.assertEqual(result.count('.'), 4)

        token = jwt.decode(result, key)
        self.assertEqual(token.claims, claims)

    def test_using_registry(self):
        key = OctKey.generate_key(128)
        value1 = jwt.encode(
            {"alg": "HS256"},
            {"sub": "a"},
            key, registry=jws.JWSRegistry()
        )
        jwt.decode(value1, key, registry=jws.JWSRegistry())
        value2 = jwt.encode(
            {"alg": "A128KW", "enc": "A128GCM"},
            {"sub": "a"},
            key, registry=jwe.JWERegistry()
        )
        jwt.decode(value2, key, registry=jwe.JWERegistry())

        self.assertRaises(
            AssertionError,
            jwt.encode,
            {"alg": "HS256"},
            {"sub": "a"},
            key, registry=jwe.JWERegistry(),
        )
        self.assertRaises(
            AssertionError,
            jwt.encode,
            {"alg": "A128KW", "enc": "A128GCM"},
            {"sub": "a"},
            key, registry=jws.JWSRegistry(),
        )
        self.assertRaises(
            AssertionError,
            jwt.decode,
            value1, key, registry=jwe.JWERegistry(),
        )
        self.assertRaises(
            AssertionError,
            jwt.decode,
            value2, key, registry=jws.JWSRegistry(),
        )



import typing as t
import copy
from .model import (
    HeaderMember,
    GeneralJSONSignature,
    FlattenedJSONSignature,
)
from .types import (
    HeaderDict,
    JSONSignatureDict,
    GeneralJSONSerialization,
    FlattenedJSONSerialization,
)
from .registry import JWSRegistry
from ..util import (
    json_b64encode,
    json_b64decode,
    urlsafe_b64encode,
    urlsafe_b64decode,
)
from ..errors import DecodeError

FindKey = t.Callable[[HeaderMember], t.Any]


def sign_general_json(
        members: t.List[HeaderDict],
        payload: bytes,
        registry: JWSRegistry,
        find_key: FindKey) -> GeneralJSONSerialization:

    payload_segment = urlsafe_b64encode(payload)
    signatures: t.List[JSONSignatureDict] = [
        __sign_member(payload_segment, HeaderMember(**member), registry, find_key)
        for member in members
    ]
    return {
        "payload": payload_segment.decode("utf-8"),
        "signatures": signatures,
    }


def sign_flattened_json(
        member: HeaderDict,
        payload: bytes,
        registry: JWSRegistry,
        find_key: FindKey) -> FlattenedJSONSerialization:
    payload_segment = urlsafe_b64encode(payload)
    signature = __sign_member(payload_segment, HeaderMember(**member), registry, find_key)
    data = {"payload": payload_segment.decode("utf-8"), **signature}
    return data  # type: ignore[return-value]


def __sign_member(
        payload_segment: bytes,
        member: HeaderMember,
        registry: JWSRegistry,
        find_key: FindKey) -> JSONSignatureDict:
    headers = member.headers()
    registry.check_header(headers)
    alg = registry.get_alg(headers["alg"])
    key = find_key(member)
    key.check_use("sig")
    alg.check_key_type(key)
    if member.protected:
        protected_segment = json_b64encode(member.protected)
    else:
        protected_segment = b""
    signing_input = b".".join([protected_segment, payload_segment])
    signature = urlsafe_b64encode(alg.sign(signing_input, key))
    rv: JSONSignatureDict = {"signature": signature.decode("utf-8")}
    if member.protected:
        rv["protected"] = protected_segment.decode("utf-8")
    if member.header:
        rv["header"] = member.header
    return rv


def extract_general_json(value: GeneralJSONSerialization) -> GeneralJSONSignature:
    payload_segment: bytes = value["payload"].encode("utf-8")
    try:
        payload = urlsafe_b64decode(payload_segment)
    except (TypeError, ValueError):
        raise DecodeError("Invalid payload")

    signatures: t.List[JSONSignatureDict] = value["signatures"]
    members = [__signature_to_member(sig) for sig in signatures]
    obj = GeneralJSONSignature(members, payload)
    obj.signatures = signatures
    obj.segments = {"payload": payload_segment}
    return obj


def extract_flattened_json(value: FlattenedJSONSerialization) -> FlattenedJSONSignature:
    payload_segment: bytes = value["payload"].encode("utf-8")
    try:
        payload = urlsafe_b64decode(payload_segment)
    except (TypeError, ValueError):
        raise DecodeError("Invalid payload")

    _sig: JSONSignatureDict = {"signature": value["signature"]}
    if "protected" in value:
        _sig["protected"] = value["protected"]
    if "header" in value:
        _sig["header"] = value["header"]

    member = __signature_to_member(_sig)
    obj = FlattenedJSONSignature(member, payload)
    obj.signature = _sig
    obj.segments = {"payload": payload_segment}
    return obj


def __signature_to_member(sig: JSONSignatureDict) -> HeaderMember:
    member = HeaderMember()
    if "protected" in sig:
        protected_segment = sig["protected"]
        member.protected = json_b64decode(protected_segment)
    if "header" in sig:
        member.header = sig["header"]
    return member


def verify_general_json(
        obj: GeneralJSONSignature,
        registry: JWSRegistry,
        find_key: FindKey) -> bool:
    payload_segment = obj.segments["payload"]
    for index, signature in enumerate(obj.signatures):
        member = obj.members[index]
        if not verify_signature(member, signature, payload_segment, registry, find_key):
            return False
    return True


def verify_flattened_json(
        obj: FlattenedJSONSignature,
        registry: JWSRegistry,
        find_key: FindKey) -> bool:
    payload_segment = obj.segments["payload"]
    assert obj.signature is not None
    return verify_signature(obj.member, obj.signature, payload_segment, registry, find_key)


def verify_signature(
        member: HeaderMember,
        signature: JSONSignatureDict,
        payload_segment: bytes,
        registry: JWSRegistry,
        find_key: FindKey) -> bool:
    headers = member.headers()
    registry.check_header(headers)
    alg = registry.get_alg(headers["alg"])
    key = find_key(member)
    key.check_use("sig")
    alg.check_key_type(key)
    if "protected" in signature:
        protected_segment = signature["protected"].encode("utf-8")
    else:
        protected_segment = b""
    sig = urlsafe_b64decode(signature["signature"].encode("utf-8"))
    signing_input = b".".join([protected_segment, payload_segment])
    return alg.verify(signing_input, sig, key)


def detach_json_content(value: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
    # https://www.rfc-editor.org/rfc/rfc7515#appendix-F
    rv = copy.deepcopy(value)  # don't alter original value
    if "payload" in rv:
        del rv["payload"]
    return rv



import typing as t
from functools import cached_property
from cryptography.hazmat.primitives.asymmetric.rsa import (
    generate_private_key,
    RSAPublicKey,
    RSAPrivateKey,
    RSAPrivateNumbers,
    RSAPublicNumbers,
    rsa_recover_prime_factors,
    rsa_crt_dmp1,
    rsa_crt_dmq1,
    rsa_crt_iqmp,
)
from cryptography.hazmat.backends import default_backend
from ..registry import KeyParameter
from ..rfc7517.models import AsymmetricKey
from ..rfc7517.pem import CryptographyBinding
from ..rfc7517.types import KeyParameters
from ..util import int_to_base64, base64_to_int


RSADictKey = t.TypedDict("RSADictKey", {
    "n": str,
    "e": str,
    "d": str,
    "p": str,
    "q": str,
    "dp": str,
    "dq": str,
    "qi": str,
}, total=False)


class RSABinding(CryptographyBinding):
    ssh_type = b"ssh-rsa"

    @staticmethod
    def import_private_key(obj: RSADictKey) -> RSAPrivateKey:
        if "oth" in obj:  # pragma: no cover
            # https://tools.ietf.org/html/rfc7518#section-6.3.2.7
            raise ValueError('"oth" is not supported yet')

        public_numbers = RSAPublicNumbers(base64_to_int(obj["e"]), base64_to_int(obj["n"]))

        if has_all_prime_factors(obj):
            numbers = RSAPrivateNumbers(
                d=base64_to_int(obj["d"]),
                p=base64_to_int(obj["p"]),
                q=base64_to_int(obj["q"]),
                dmp1=base64_to_int(obj["dp"]),
                dmq1=base64_to_int(obj["dq"]),
                iqmp=base64_to_int(obj["qi"]),
                public_numbers=public_numbers,
            )
        else:
            d = base64_to_int(obj["d"])
            p, q = rsa_recover_prime_factors(public_numbers.n, d, public_numbers.e)
            numbers = RSAPrivateNumbers(
                d=d,
                p=p,
                q=q,
                dmp1=rsa_crt_dmp1(d, p),
                dmq1=rsa_crt_dmq1(d, q),
                iqmp=rsa_crt_iqmp(p, q),
                public_numbers=public_numbers,
            )

        return numbers.private_key(default_backend())

    @staticmethod
    def export_private_key(key: RSAPrivateKey) -> RSADictKey:
        numbers = key.private_numbers()
        return {
            "n": int_to_base64(numbers.public_numbers.n),
            "e": int_to_base64(numbers.public_numbers.e),
            "d": int_to_base64(numbers.d),
            "p": int_to_base64(numbers.p),
            "q": int_to_base64(numbers.q),
            "dp": int_to_base64(numbers.dmp1),
            "dq": int_to_base64(numbers.dmq1),
            "qi": int_to_base64(numbers.iqmp),
        }

    @staticmethod
    def import_public_key(obj: RSADictKey) -> RSAPublicKey:
        numbers = RSAPublicNumbers(base64_to_int(obj["e"]), base64_to_int(obj["n"]))
        return numbers.public_key(default_backend())

    @staticmethod
    def export_public_key(key: RSAPublicKey) -> t.Dict[str, str]:
        numbers = key.public_numbers()
        return {"n": int_to_base64(numbers.n), "e": int_to_base64(numbers.e)}


class RSAKey(AsymmetricKey[RSAPrivateKey, RSAPublicKey]):
    key_type = "RSA"
    #: Registry definition for RSA Key
    #: https://www.rfc-editor.org/rfc/rfc7518#section-6.3
    value_registry = {
        "n": KeyParameter("Modulus", "str", private=False, required=True),
        "e": KeyParameter("Exponent", "str", private=False, required=True),
        "d": KeyParameter("Private Exponent", "str", private=True, required=False),
        "p": KeyParameter("First Prime Factor", "str", private=True, required=False),
        "q": KeyParameter("Second Prime Factor", "str", private=True, required=False),
        "dp": KeyParameter("First Factor CRT Exponent", "str", private=True, required=False),
        "dq": KeyParameter("Second Factor CRT Exponent", "str", private=True, required=False),
        "qi": KeyParameter("First CRT Coefficient", "str", private=True, required=False),
        "oth": KeyParameter("Other Primes Info", "none", private=True, required=False),
    }
    binding = RSABinding

    @property
    def is_private(self) -> bool:
        return isinstance(self.raw_value, RSAPrivateKey)

    @cached_property
    def public_key(self) -> RSAPublicKey:
        if isinstance(self.raw_value, RSAPrivateKey):
            return self.raw_value.public_key()
        return self.raw_value

    @property
    def private_key(self) -> t.Optional[RSAPrivateKey]:
        if isinstance(self.raw_value, RSAPrivateKey):
            return self.raw_value
        return None

    @classmethod
    def generate_key(
            cls,
            key_size: int = 2048,
            parameters: t.Optional[KeyParameters] = None,
            private: bool = True,
            auto_kid: bool = False) -> "RSAKey":
        """Generate a ``RSAKey`` with the given bit size (not bytes).

        :param key_size: size in bit
        :param parameters: extra parameter in JWK
        :param private: generate a private key or public key
        :param auto_kid: add ``kid`` automatically
        """
        if key_size < 512:
            raise ValueError("key_size must not be less than 512")
        if key_size % 8 != 0:
            raise ValueError("Invalid key_size for RSAKey")
        raw_key = generate_private_key(
            public_exponent=65537,
            key_size=key_size,
            backend=default_backend(),
        )
        if private:
            key = cls(raw_key, raw_key, parameters)
        else:
            pub_key = raw_key.public_key()
            key = cls(pub_key, pub_key, parameters)
        if auto_kid:
            key.ensure_kid()
        return key


def has_all_prime_factors(obj: RSADictKey) -> bool:
    props = ["p", "q", "dp", "dq", "qi"]
    props_found = [prop in obj for prop in props]
    if all(props_found):
        return True

    if any(props_found):
        raise ValueError("RSA key must include all parameters if any are present besides d")

    return False



from unittest import TestCase
from joserfc.jwk import ECKey
from joserfc.errors import InvalidExchangeKeyError
from tests.keys import read_key


class TestECKey(TestCase):
    def test_exchange_derive_key(self):
        key1 = ECKey.generate_key("P-256")
        key2 = ECKey.generate_key("P-384")
        self.assertRaises(InvalidExchangeKeyError, key1.exchange_derive_key, key2)

        key3 = ECKey.generate_key("P-256", private=False)
        self.assertRaises(InvalidExchangeKeyError, key3.exchange_derive_key, key1)

    def run_import_key(self, name):
        public_pem = read_key(f"ec-{name}-public.pem")
        key1 = ECKey.import_key(public_pem)
        self.assertFalse(key1.is_private)

        private_pem = read_key(f"ec-{name}-private.pem")
        key2 = ECKey.import_key(private_pem)
        self.assertTrue(key2.is_private)

        # public key match
        self.assertEqual(
            key2.as_bytes(private=False),
            key1.as_bytes(private=False),
        )

        self.assertNotIn("d", key1.dict_value)
        self.assertIn("d", key2.dict_value)

    def test_import_p256_key(self):
        self.run_import_key("p256")

    def test_import_p384_key(self):
        self.run_import_key("p384")

    def test_import_p512_key(self):
        self.run_import_key("p512")

    def test_import_secp256k1_key(self):
        self.run_import_key("secp256k1")

    def test_generate_key(self):
        self.assertRaises(ValueError, ECKey.generate_key, "Invalid")

        key = ECKey.generate_key(private=True)
        self.assertTrue(key.is_private)
        self.assertIsNone(key.kid)

        key = ECKey.generate_key(private=False)
        self.assertFalse(key.is_private)
        self.assertIsNone(key.private_key)

        key = ECKey.generate_key(auto_kid=True)
        self.assertIsNotNone(key.kid)

    def test_import_from_der_bytes(self):
        origin_key = ECKey.generate_key()
        value1 = origin_key.as_der()
        value2 = origin_key.as_der(private=False)

        key1 = ECKey.import_key(value1)
        key2 = ECKey.import_key(value2)
        self.assertEqual(value1, key1.as_der())
        self.assertEqual(value2, key1.as_der(private=False))
        self.assertEqual(value2, key2.as_der())

    def test_output_with_password(self):
        key = ECKey.import_key(read_key("ec-p256-private.pem"))
        pem = key.as_pem(password="secret")
        self.assertRaises(
            TypeError,
            ECKey.import_key,
            pem
        )
        key2 = ECKey.import_key(pem, password="secret")
        self.assertEqual(key.as_dict(), key2.as_dict())



from unittest import TestCase
from joserfc.jws import register_key_set
from joserfc.jwk import JWKRegistry, guess_key
from joserfc.jwk import KeySet, OctKey, RSAKey
from joserfc.errors import (
    UnsupportedKeyAlgorithmError,
    UnsupportedKeyUseError,
    UnsupportedKeyOperationError,
)

register_key_set()


class Guest:
    def __init__(self):
        self._headers = {}

    def headers(self):
        return self._headers

    def set_kid(self, kid):
        self._headers["kid"] = kid


class TestKeyMethods(TestCase):
    def test_guess_str_key(self):
        key = guess_key("key", Guest())
        self.assertIsInstance(key, OctKey)

    def test_guess_bytes_key(self):
        key = guess_key(b"key", Guest())
        self.assertIsInstance(key, OctKey)

    def test_guess_callable_key(self):
        oct_key = OctKey.generate_key(parameters={'kid': '1'})
        rsa_key = RSAKey.generate_key(parameters={'kid': '2'})

        def key_func1(obj):
            return "key"

        def key_func2(obj):
            return rsa_key

        def key_func3(obj):
            return KeySet([oct_key, rsa_key])

        key = guess_key(key_func1, Guest())
        self.assertIsInstance(key, OctKey)

        key = guess_key(key_func2, Guest())
        self.assertIsInstance(key, RSAKey)

        guest = Guest()
        guest.set_kid('2')
        key = guess_key(key_func3, guest)
        self.assertIsInstance(key, RSAKey)

    def test_guess_key_set(self):
        key_set = KeySet([OctKey.generate_key(), RSAKey.generate_key()])
        guest = Guest()
        guest._headers["alg"] = "HS256"
        self.assertRaises(ValueError, guess_key, key_set, guest)
        key = guess_key(key_set, guest, True)
        self.assertIsInstance(key, OctKey)
        key = guess_key(key_set, guest)

        guest = Guest()
        guest._headers["alg"] = "RS256"
        self.assertRaises(ValueError, guess_key, key_set, guest)
        key = guess_key(key_set, guest, True)
        self.assertIsInstance(key, RSAKey)

        guest = Guest()
        guest._headers["alg"] = "ES256"
        self.assertRaises(ValueError, guess_key, key_set, guest, True)

    def test_invalid_key(self):
        self.assertRaises(ValueError, guess_key, {}, Guest())

    def test_import_key(self):
        # test bytes
        key = JWKRegistry.import_key(b"secret", "oct")
        self.assertIsInstance(key, OctKey)

        # test string
        key = JWKRegistry.import_key("secret", "oct")
        self.assertIsInstance(key, OctKey)

        # test dict
        data = key.as_dict()
        key = JWKRegistry.import_key(data)
        self.assertIsInstance(key, OctKey)

        self.assertRaises(ValueError, JWKRegistry.import_key, "secret", "invalid")

    def test_generate_key(self):
        key = JWKRegistry.generate_key("oct", 8)
        self.assertIsInstance(key, OctKey)
        self.assertRaises(ValueError, JWKRegistry.generate_key, "invalid", 8)

    def test_check_use(self):
        key = OctKey.import_key("secret", {"use": "sig"})
        key.check_use("sig")
        self.assertRaises(
            UnsupportedKeyUseError,
            key.check_use,
            "enc"
        )
        self.assertRaises(
            UnsupportedKeyUseError,
            key.check_use,
            "invalid"
        )

    def test_check_alg(self):
        key = OctKey.import_key("secret", {"alg": "HS256"})
        key.check_alg("HS256")
        self.assertRaises(
            UnsupportedKeyAlgorithmError,
            key.check_alg,
            "RS256"
        )

    def test_alg_property(self):
        key = OctKey.import_key("secret")
        self.assertIsNone(key.alg)

        key = OctKey.import_key("secret", {"alg": "HS256"})
        self.assertEqual(key.alg, "HS256")

    def test_check_ops(self):
        key = OctKey.import_key("secret", {"key_ops": ["sign", "verify"]})
        key.check_key_op("sign")
        self.assertRaises(
            UnsupportedKeyOperationError,
            key.check_key_op,
            "wrapKey"
        )
        self.assertRaises(
            UnsupportedKeyOperationError,
            key.check_key_op,
            "invalid"
        )
        key = RSAKey.generate_key(private=False)
        self.assertRaises(
            UnsupportedKeyOperationError,
            key.check_key_op,
            "sign"
        )

    def test_import_without_kty(self):
        self.assertRaises(ValueError, JWKRegistry.import_key, {})



from unittest import TestCase
from joserfc.jwe import JWERegistry, encrypt_compact, decrypt_compact, CompactEncryption
from joserfc.jwk import RSAKey, ECKey, OctKey, OKPKey, KeySet
from joserfc.rfc7518.jwe_encs import JWE_ENC_MODELS
from joserfc.errors import (
    InvalidKeyLengthError,
    MissingAlgorithmError,
    MissingEncryptionError,
    DecodeError,
)
from joserfc.util import json_b64encode
from tests.base import load_key


class TestJWECompact(TestCase):
    def run_case(self, alg: str, enc: str, private_key, public_key):
        protected = {"alg": alg, "enc": enc}
        payload = b'hello'
        result = encrypt_compact(
            protected, payload, public_key,
            algorithms=[alg, enc],
        )
        self.assertEqual(result.count('.'), 4)

        obj = decrypt_compact(
            result, private_key,
            algorithms=[alg, enc],
        )
        self.assertEqual(obj.plaintext, payload)

    def run_cases(self, names, private_key, public_key):
        for alg in names:
            for enc in JWE_ENC_MODELS:
                self.run_case(alg, enc.name, private_key, public_key)

    def test_RSA_alg(self):
        private_key: RSAKey = load_key('rsa-openssl-private.pem')
        public_key: RSAKey = load_key('rsa-openssl-public.pem')
        algs = ['RSA1_5', 'RSA-OAEP', 'RSA-OAEP-256']
        self.run_cases(algs, private_key, public_key)

        protected = {"alg": "RSA-OAEP", "enc": "A128CBC-HS256"}
        value = encrypt_compact(protected, "i", private_key)
        key2 = RSAKey.generate_key()
        self.assertRaises(
            DecodeError,
            decrypt_compact,
            value, key2
        )

    def test_ECDH_ES_with_EC_key(self):
        algs = ['ECDH-ES', 'ECDH-ES+A128KW', 'ECDH-ES+A192KW', 'ECDH-ES+A256KW']
        for size in [256, 384, 512]:
            private_key: ECKey = load_key(f'ec-p{size}-private.pem')
            public_key: ECKey = load_key(f'ec-p{size}-public.pem')
            self.run_cases(algs, private_key, public_key)

        key1 = ECKey.generate_key("P-256")
        key2 = ECKey.generate_key("P-256")
        key3 = ECKey.generate_key("P-521")
        for alg in ["ECDH-ES", "ECDH-ES+A128KW"]:
            for enc in ["A128CBC-HS256", "A128GCM"]:
                protected = {"alg": alg, "enc": enc}
                value = encrypt_compact(protected, "i", key1)
                self.assertRaises(
                    DecodeError,
                    decrypt_compact,
                    value, key2,
                )
                self.assertRaises(
                    DecodeError,
                    decrypt_compact,
                    value, key3,
                )

    def test_ECDH_ES_with_OKP_key(self):
        key1 = OKPKey.generate_key("X25519")
        key2 = OKPKey.generate_key("X448")
        for alg in ['ECDH-ES', 'ECDH-ES+A128KW']:
            for enc in ["A128CBC-HS256", "A128GCM"]:
                protected = {"alg": alg, "enc": enc}
                value = encrypt_compact(protected, "i", key1)
                obj = decrypt_compact(value, key1)
                self.assertEqual(obj.protected, protected)
                self.assertRaises(
                    DecodeError,
                    decrypt_compact,
                    value, key2,
                )

                value = encrypt_compact(protected, "i", key2)
                obj = decrypt_compact(value, key2)
                self.assertEqual(obj.protected, protected)
                self.assertRaises(
                    DecodeError,
                    decrypt_compact,
                    value, key1,
                )

    def test_dir_alg(self):
        self.assertRaises(InvalidKeyLengthError, encrypt_compact, {"alg": "dir", "enc": "A128GCM"}, b"j", "secret")
        for enc in JWE_ENC_MODELS:
            key = OctKey.generate_key(enc.cek_size)
            self.run_case('dir', enc.name, key, key)

    def test_AESGCM_alg(self):
        for size in [128, 192, 256]:
            key = OctKey.generate_key(size)
            self.run_cases([f"A{size}GCMKW"], key, key)

        key1 = OctKey.generate_key(128)
        key2 = OctKey.generate_key(128)
        protected = {"alg": "A128GCMKW", "enc": "A128CBC-HS256"}
        algorithms = ["A128GCMKW", "A128CBC-HS256"]
        value = encrypt_compact(protected, "i", key1, algorithms=algorithms)
        self.assertRaises(
            DecodeError,
            decrypt_compact,
            value, key2, algorithms=algorithms,
        )

    def test_PBES2HS_alg(self):
        algs = {
            "PBES2-HS256+A128KW": 128,
            "PBES2-HS384+A192KW": 192,
            "PBES2-HS512+A256KW": 256,
        }
        for alg in algs:
            key = OctKey.generate_key(algs[alg])
            self.run_cases([alg], key, key)

        key1 = OctKey.generate_key(128)
        key2 = OctKey.generate_key(128)
        protected = {"alg": "PBES2-HS256+A128KW", "enc": "A128CBC-HS256"}
        algorithms = ["PBES2-HS256+A128KW", "A128CBC-HS256"]
        value = encrypt_compact(protected, "i", key1, algorithms=algorithms)
        self.assertRaises(
            DecodeError,
            decrypt_compact,
            value, key2, algorithms=algorithms,
        )

    def test_PBES2HS_with_header(self):
        key = OctKey.generate_key(128)
        protected = {
            "alg": "PBES2-HS256+A128KW",
            "enc": "A128CBC-HS256",
            "p2s": "QoGrcBpns_cLWCQPEVuA-g",
            "p2c": 1024,
        }
        registry = JWERegistry(algorithms=["PBES2-HS256+A128KW", "A128CBC-HS256"])
        value1 = encrypt_compact(protected, b"i", key, registry=registry)
        obj1 = decrypt_compact(value1, key, registry=registry)
        self.assertIn("p2c", obj1.headers())
        self.assertEqual(obj1.headers()["p2c"], 1024)

        # invalid type
        protected["p2c"] = "1024"
        self.assertRaises(
            ValueError,
            encrypt_compact,
            protected, b"i", key,
            registry=registry,
        )

    def test_with_zip_header(self):
        private_key: RSAKey = load_key('rsa-openssl-private.pem')
        public_key: RSAKey = load_key('rsa-openssl-public.pem')
        protected = {"alg": "RSA-OAEP", "enc": "A128CBC-HS256", "zip": "DEF"}
        plaintext = b"hello"
        result = encrypt_compact(protected, plaintext, public_key)
        obj = decrypt_compact(result, private_key)
        self.assertEqual(obj.plaintext, plaintext)

    def test_invalid_compact_data(self):
        private_key: RSAKey = load_key('rsa-openssl-private.pem')
        value = b"a.b.c.d.e.f.g"
        self.assertRaises(ValueError, decrypt_compact, value, private_key)
        value = b"a.b.c.d.e"
        self.assertRaises(DecodeError, decrypt_compact, value, private_key)

        value = json_b64encode({"enc": "A128CBC-HS256"}) + b".b.c.d.e"
        self.assertRaises(MissingAlgorithmError, decrypt_compact, value, private_key)

        value = json_b64encode({"alg": "RSA-OAEP"}) + b".b.c.d.e"
        self.assertRaises(MissingEncryptionError, decrypt_compact, value, private_key)

    def test_with_key_set(self):
        keys = KeySet([
            OctKey.generate_key(),
            OctKey.generate_key(),
            RSAKey.generate_key(),
        ])
        protected = {"alg": "RSA-OAEP", "enc": "A128CBC-HS256"}
        value = encrypt_compact(protected, b"foo", keys)
        obj = decrypt_compact(value, keys)
        self.assertEqual(obj.plaintext, b"foo")

    def test_compact_encryption(self):
        key: RSAKey = load_key('rsa-openssl-private.pem')
        protected = {"alg": "RSA-OAEP", "enc": "A128CBC-HS256"}
        obj = CompactEncryption(protected, b"")
        self.assertEqual(obj.recipients, [])
        obj.attach_recipient(key, {"kid": "foo"})
        self.assertEqual(obj.protected["kid"], "foo")



from unittest import TestCase
from joserfc.jwe import (
    encrypt_compact,
    decrypt_compact,
    JWERegistry,
)
from joserfc.jwk import OctKey
from joserfc.drafts.jwe_chacha20 import register_chaha20_poly1305

register_chaha20_poly1305()
chacha_registry = JWERegistry(algorithms=["dir", "C20P", "XC20P"])


class TestChaCha20(TestCase):
    def run_test_dir(self, enc: str):
        key = OctKey.generate_key(256)
        protected = {"alg": "dir", "enc": enc}
        encrypted_text = encrypt_compact(protected, b'hello', key, registry=chacha_registry)
        self.assertEqual(encrypted_text.count("."), 4)
        obj = decrypt_compact(encrypted_text, key, registry=chacha_registry)
        self.assertEqual(obj.plaintext, b'hello')

        key2 = OctKey.generate_key(256)
        self.assertRaises(ValueError, decrypt_compact, encrypted_text, key2, registry=chacha_registry)

    def test_dir_C20P(self):
        self.run_test_dir("C20P")

    def test_dir_XC20P(self):
        self.run_test_dir("XC20P")

    def test_xc20p_content_encryption_decryption(self):
        # https://datatracker.ietf.org/doc/html/draft-irtf-cfrg-xchacha-03#appendix-A.3.1
        plaintext = bytes.fromhex(
            '4c616469657320616e642047656e746c656d656e206f662074686520636c6173'
            '73206f66202739393a204966204920636f756c64206f6666657220796f75206f'
            '6e6c79206f6e652074697020666f7220746865206675747572652c2073756e73'
            '637265656e20776f756c642062652069742e'
        )

        cek = bytes.fromhex('808182838485868788898a8b8c8d8e8f909192939495969798999a9b9c9d9e9f')
        iv = bytes.fromhex('404142434445464748494a4b4c4d4e4f5051525354555657')
        aad = bytes.fromhex('50515253c0c1c2c3c4c5c6c7')

        XC20P = chacha_registry.get_enc("XC20P")
        ciphertext, tag = XC20P.encrypt(plaintext, cek, iv, aad)
        self.assertEqual(
            ciphertext,
            bytes.fromhex(
                'bd6d179d3e83d43b9576579493c0e939572a1700252bfaccbed2902c21396cbb'
                '731c7f1b0b4aa6440bf3a82f4eda7e39ae64c6708c54c216cb96b72e1213b452'
                '2f8c9ba40db5d945b11b69b982c1bb9e3f3fac2bc369488f76b2383565d3fff9'
                '21f9664c97637da9768812f615c68b13b52e'
            )
        )

        self.assertEqual(tag, bytes.fromhex('c0875924c1c7987947deafd8780acf49'))

        result = XC20P.decrypt(ciphertext, tag, cek, iv, aad)
        self.assertEqual(plaintext, result)



import typing as t
from ._keys import (
    JWKRegistry,
    KeySet,
    OctKey,
    RSAKey,
    ECKey,
    OKPKey,
    Key,
)
from .rfc8812 import register_secp256k1
from .registry import Header


__all__ = [
    "JWKRegistry",
    "Key",
    "KeyCallable",
    "KeyFlexible",
    "OctKey",
    "RSAKey",
    "ECKey",
    "OKPKey",
    "KeySet",
    "guess_key",
]

register_secp256k1()


class GuestProtocol(t.Protocol):  # pragma: no cover
    def headers(self) -> Header:
        ...

    def set_kid(self, kid: str) -> None:
        ...


KeyBase = t.Union[str, bytes, Key, KeySet]
KeyCallable = t.Callable[[GuestProtocol], KeyBase]
KeyFlexible = t.Union[KeyBase, KeyCallable]


def guess_key(key: KeyFlexible, obj: GuestProtocol, use_random: bool = False) -> Key:
    """Guess key from a various sources.

    :param key: a very flexible key
    :param obj: a protocol that has ``headers`` and ``set_kid`` methods
    :param use_random: pick a random key from key set
    """

    _norm_key: t.Union[Key, KeySet]
    if callable(key):
        _norm_key = _normalize_key(key(obj))
    else:
        _norm_key = _normalize_key(key)

    rv_key: Key
    if isinstance(_norm_key, KeySet):
        headers = obj.headers()
        kid = headers.get("kid")
        if not kid and use_random:
            # choose one key by random
            rv_key = _norm_key.pick_random_key(headers["alg"])  # type: ignore[assignment]
            if rv_key is None:
                raise ValueError("Invalid key")
            rv_key.ensure_kid()
            assert rv_key.kid is not None  # for mypy
            obj.set_kid(rv_key.kid)
        else:
            rv_key = _norm_key.get_by_kid(kid)
    elif isinstance(_norm_key, (OctKey, RSAKey, ECKey, OKPKey)):
        rv_key = _norm_key
    else:
        raise ValueError("Invalid key")
    return rv_key


def _normalize_key(key: KeyBase) -> t.Union[Key, KeySet]:
    if isinstance(key, (str, bytes)):
        return OctKey.import_key(key)
    return key
